@@ -45,9 +45,13 @@ class TokenizerConstructor:
         self.start_token = self.special_tokens.index(start_token_string) if start_token_string is not None else None
         self.end_token = self.special_tokens.index(end_token_string) if end_token_string is not None else None
         self.pad_token = self.special_tokens.index(pad_token_string) if pad_token_string is not None else None
-        self.pad_token_string = pad_token_string
         self.new_line_token = self.special_tokens.index(new_line_token_string) if new_line_token_string is not None else None
 
+        self.start_token_string = start_token_string
+        self.end_token_string = end_token_string
+        self.pad_token_string = pad_token_string
+        self.new_line_token_string = new_line_token_string
+
         if tokenizer_type == "BPE":
             self.tokenizer_type = tokenizers.Tokenizer(tokenizers.models.BPE(unk_token=unknown_token_string))
             self.trainer = tokenizers.trainers.BpeTrainer(special_tokens=self.special_tokens, min_frequency=min_frequency, vocab_size=vocab_size)
@@ -136,10 +140,12 @@ class TokenizerConstructor:
         '''
         if max_length is not None:
             self.tokenizer_type.enable_truncation(max_length=max_length)
-            self.tokenizer_type.enable_padding(pad_id=self.pad_token, pad_token=self.pad_token_string, length=max_length)
+            if self.pad_token is not None:
+                self.tokenizer_type.enable_padding(pad_id=self.pad_token, pad_token=self.pad_token_string, length=max_length)
         out = [row.ids for row in self.tokenizer_type.encode_batch(inp)]
         self.tokenizer_type.no_truncation()
-        self.tokenizer_type.enable_padding(pad_id=self.pad_token, pad_token=self.pad_token_string)
+        if self.pad_token is not None:
+            self.tokenizer_type.enable_padding(pad_id=self.pad_token, pad_token=self.pad_token_string)
         return out
     
     def decode(self, inp:list[int]) -> str:
@@ -215,6 +221,7 @@ class DataProcessor:
         if isinstance(dec_data, str):
             dec_data = [dec_data]
         dec_data_length = len(dec_data)
+        dec_data = pre_process_data(dec_data, self.dec_tokenizer.start_token_string, self.dec_tokenizer.end_token_string)
 
         if enc_data is not None:
             if self.enc_tokenizer is None:
@@ -223,6 +230,7 @@ class DataProcessor:
             enc_data_length = len(enc_data)
             if dec_data_length != enc_data_length:
                 raise Exception(f"decoder data and encoder data lengths do not match. decoder_data_length is {dec_data_length}, encoder_data_length is {enc_data_length}")
+            enc_data = pre_process_data(enc_data, self.enc_tokenizer.start_token_string, self.enc_tokenizer.end_token_string)
 
         print("processing data")
         dec_out_list = self.dec_tokenizer.encode_batch(dec_data, max_length=dec_max_block_size)
@@ -281,14 +289,14 @@ def get_batch(data:torch.Tensor,
             random_point = get_valid_samples(random_samples, masks, block_size)
         else:
             random_point = torch.randint(data.shape[1] - block_size, (batch_size,))
-        batch_in = torch.stack([data[random_samples[i]][random_point[i]:random_point[i]+block_size-int(get_offset)] for i in range(batch_size)])
-        masks_in = torch.stack([masks[random_samples[i]][random_point[i]:random_point[i]+block_size-int(get_offset)] for i in range(batch_size)]) if masks is not None else None
-        batch_out = torch.stack([data[random_samples[i]][1+random_point[i]:random_point[i]+block_size] for i in range(batch_size)]) if get_offset else None
+        batch_in = safe_stack([data[random_samples[i]][random_point[i]:random_point[i]+block_size-int(get_offset)] for i in range(batch_size)])
+        masks_in = safe_stack([masks[random_samples[i]][random_point[i]:random_point[i]+block_size-int(get_offset)] for i in range(batch_size)]) if masks is not None else None
+        batch_out = safe_stack([data[random_samples[i]][1+random_point[i]:random_point[i]+block_size] for i in range(batch_size)]) if get_offset else None
     else:
         block_size = data.shape[1]
-        batch_in = torch.stack([data[row_num][:block_size-int(get_offset)] for row_num in random_samples])
-        masks_in = torch.stack([masks[row_num][:block_size-int(get_offset)] for row_num in random_samples]) if masks is not None else None
-        batch_out = torch.stack([data[row_num][1:block_size] for row_num in random_samples]) if get_offset else None
+        batch_in = safe_stack([data[row_num][:block_size-int(get_offset)] for row_num in random_samples])
+        masks_in = safe_stack([masks[row_num][:block_size-int(get_offset)] for row_num in random_samples]) if masks is not None else None
+        batch_out = safe_stack([data[row_num][1:block_size] for row_num in random_samples]) if get_offset else None
 
     return batch_in, batch_out, masks_in
 
@@ -385,6 +393,7 @@ class SelfAttention(nn.Module):
 
         k = self.key(k)
         q = self.query(q)
+        v = self.value(v)
 
         wei = q @ k.transpose(-2,-1) * k.shape[-1]**-0.5
         if self.triangle_mask and self.block_size >= 0:
@@ -394,7 +403,6 @@ class SelfAttention(nn.Module):
         wei = F.softmax(wei, dim=-1)
         wei = self.dropout(wei)
 
-        v = self.value(v)
         out = wei @ v
         return out
 
@@ -559,7 +567,6 @@ class RoboConstructor(nn.Module):
                  enc_vocab_size:int=None,
                  enc_block_size:int=None,
                  enc_expansion_factor:int=4,
-                 enc_positional_encoding:bool=True,
                  dropout:float=0.1,
                  device:str=None
                  ) -> None:
@@ -580,7 +587,6 @@ class RoboConstructor(nn.Module):
         self.dec_positional_embedding_table = nn.Embedding(dec_block_size, n_embed)
 
         if enc_n_blocks != 0:
-            self.enc_positional_encoding = enc_positional_encoding
             self.enc_n_blocks = enc_n_blocks
             self.enc_n_head = enc_n_head
             self.enc_expansion_factor = enc_expansion_factor
@@ -588,8 +594,7 @@ class RoboConstructor(nn.Module):
             self.enc_block_size = enc_block_size
             self.cross_attention = True
             self.enc_token_embedding_table = nn.Embedding(enc_vocab_size, n_embed)
-            if enc_positional_encoding:
-                self.enc_positional_embedding_table = nn.Embedding(enc_block_size, n_embed)
+            self.enc_positional_embedding_table = nn.Embedding(enc_block_size, n_embed)
             self.encoder_blocks = MySequential(*[EncoderBlock(n_embed, enc_n_head, enc_expansion_factor, dropout=dropout) for _ in range(enc_n_blocks)])
         else:
             self.cross_attention = False
@@ -632,11 +637,9 @@ class RoboConstructor(nn.Module):
 
         if self.cross_attention:
             enc_tok_emb = self.enc_token_embedding_table(enc_in)
-            if self.enc_positional_encoding:
-                enc_pos_emb = self.enc_positional_embedding_table(torch.arange(enc_T, device=self.device))
-                enc_x = enc_tok_emb + enc_pos_emb
-            else:
-                enc_x = enc_tok_emb
+            enc_pos_emb = self.enc_positional_embedding_table(torch.arange(enc_T, device=self.device))
+            enc_x = enc_tok_emb + enc_pos_emb
+
 
             enc_out, enc_mask = self.encoder_blocks(enc_x, enc_mask)
         else:
@@ -694,7 +697,8 @@ class RoboConstructor(nn.Module):
               pad_token:int=None,
               dec_tokenizer:TokenizerConstructor=None,
               save_path:str=None,
-              label_smoothing:float=0.1
+              label_smoothing:float=0.1,
+              optimizer_state_dict_path:str=None
               ) -> None:
         '''
         trains the RoboConstructor instance transformer.
@@ -741,6 +745,13 @@ class RoboConstructor(nn.Module):
             loss_fn = nn.CrossEntropyLoss(label_smoothing=label_smoothing).to(self.device)
         print(sum(p.numel() for p in self.parameters())/1e6, "M parameters")
         optimizer = torch.optim.AdamW(self.parameters(), lr=learning_rate)
+        if optimizer_state_dict_path is not None:
+            opt_path = os.path.join(optimizer_state_dict_path, "opt.pt")
+            if os.path.isfile(opt_path):
+                optimizer.load_state_dict(torch.load(opt_path))
+                for param_group in optimizer.param_groups:
+                    param_group["lr"] = learning_rate
+                    
         @torch.no_grad()
         def estimate_loss() -> dict:
             out = {}
@@ -748,13 +759,13 @@ class RoboConstructor(nn.Module):
             losses = torch.zeros(eval_iters)
             for k in range(eval_iters):
                 dec_x, dec_y, dec_mask, enc_x, enc_mask = self.prep_data(batch_size, dec_training_data, dec_masks=dec_training_masks_data, dec_block_size=self.dec_block_size, enc_data=enc_training_data, enc_masks=enc_training_masks_data, enc_block_size=self.enc_block_size)
-                proj_output = self.forward(dec_x, dec_mask, enc_x, enc_mask)
+                proj_output = self(dec_x, dec_mask, enc_x, enc_mask)
                 losses[k] = loss_fn(proj_output.view(-1, self.dec_vocab_size), dec_y.view(-1))
             out["train"] = losses.mean()
             if dec_eval_data is not None:
                 for k in range(eval_iters):
                     dec_x, dec_y, dec_mask, enc_x, enc_mask = self.prep_data(batch_size, dec_eval_data, dec_masks=dec_eval_masks_data, dec_block_size=self.dec_block_size, enc_data=enc_eval_data, enc_masks=enc_eval_masks_data, enc_block_size=self.enc_block_size)
-                    proj_output = self.forward(dec_x, dec_mask, enc_x, enc_mask)
+                    proj_output = self(dec_x, dec_mask, enc_x, enc_mask)
                     losses[k] = loss_fn(proj_output.view(-1, self.dec_vocab_size), dec_y.view(-1))
                 out["eval"] = losses.mean()
             else:
@@ -769,9 +780,11 @@ class RoboConstructor(nn.Module):
                 print(f"step {iter}: train loss {losses['train']:.4f}, eval loss {losses['eval']:.4f}")
                 if save_path is not None:
                     save_component(self, save_path=save_path)
+                if optimizer_state_dict_path is not None:
+                    torch.save(optimizer.state_dict(), os.path.join(optimizer_state_dict_path, "opt.pt"))
 
             dec_x, dec_y, dec_mask, enc_x, enc_mask = self.prep_data(batch_size, dec_training_data, dec_masks=dec_training_masks_data, dec_block_size=self.dec_block_size, enc_data=enc_training_data, enc_masks=enc_training_masks_data, enc_block_size=self.enc_block_size)
-            proj_output = self.forward(dec_x, dec_mask, enc_x, enc_mask)
+            proj_output = self(dec_x, dec_mask, enc_x, enc_mask)
             loss = loss_fn(proj_output.view(-1, self.dec_vocab_size), dec_y.view(-1))
             loss.backward()
             optimizer.step()
